# Copyright 2004 Tom Rothamel <pytom@bishoujo.us>
# 
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
# 
# 
# This module is intended to be used as a singleton object.
# It's purpose is to store in one global all of the data that would
# be to annoying to lug around otherwise. 
#
# Many modules will probablt want to import this using a command like:
#
# import renpy.game as game
#
# These modules will then be able to access the various globals defined
# in this module as fields on game.

import renpy.script
import renpy.execution
import renpy.python
import renpy.config

# The basepath. Everything that is loaded is relative to this, at least
# until we get around to implementing archive files.
basepath = None

# A Script object, giving the script of the currently executing game.
script = None

# The store is where Ren'Py python results are stored. We first need
# to import in the module, and then we use the module's dictionary
# directly.
store = None

# A stack of execution contexts.
contexts = [ ]

# The interface that the game uses to interact with the user.
interface = None

# Are we still running init blocks?
init_phase = True

# The RollbackLog that keeps track of changes to the game state
# and to the store.
log = None

# Some useful additional information about program execution that
# can be added to the exception.
exception_info = ''

def context(index=-1):
    """
    Return the current execution context, or the context at the
    given index if one is specified.
    """

    return contexts[index]

class RestartException(Exception):
    """
    This class will be used to convey to the system that the context has
    been changed, and therefore execution needs to be restarted.
    """

class QuitException(Exception):
    """
    An exception of this class will let us force a safe quit, from
    anywhere in the program. Do not pass go, do not collect $200.
    """

def call_in_new_context(label):
    """
    This code creates a new context, and starts executing code from
    that label in the new context. Rollback is disabled in the
    new context. (Actually, it will just bring you back to the
    real context.)
    """

    context = renpy.execution.Context(False, contexts[-1])
    contexts.append(context)

    context.goto_label(label)
    context.run()

    contexts.pop()
    interface.redraw(0)

    
    
    
def main(basepath_):
    global script
    global contexts
    global interface
    global basepath
    global store
    global init_phase
    global log

    basepath = basepath_

    # Finish loading the config.
    renpy.config.finish()

    # Load the script.
    script = renpy.script.load_script(basepath)

    # Initialize the store.
    import renpy.store as store
    store.store = store
    store = vars(renpy.store)

    # Initialize the log.
    log = renpy.python.RollbackLog()

    # Run init code in its own context. (Don't log.)
    contexts = [ renpy.execution.Context(False) ]

    # Run the init code.
    for i in script.initcode:
        context().run(i)

    # We're done running the initialization code.
    init_phase = False

    # Switch contexts, begin logging.
    contexts = [ renpy.execution.Context(True) ]

    # Initialize graphics.
    import renpy.display.core as dcore
    interface = dcore.Interface()

    # TODO: Make a clean copy of the store, for saving.

    if script.has_label("_start"):
        context().goto_label('_start')
    else:
        context().goto_label('start')

    while True:

        # We first try running the script.
        try:
            context().run()
            break

        # We get this when the context has changed, and so we go and
        # start running from the new context.
        except RestartException, e:
            pass

        except QuitException, e:
            break

    # And, we're done.
